package FDN::Sql::MultiDBC;

use strict;
use warnings;
use base 'Exporter';

use FDN::Common::Logger;
use FDN::Sql::Basic;
use FDN::Sql::DB_Access;
use FDN::Sql::DBC qw(do_disconnect);
use FDN::Sql::Structure qw(read_config read_files structure create switch_option set_quiet set_output purge);


our @EXPORT = qw(
	do_connect
	do_disconnect
	reset_multi_dbh
	switch_multi_dbh
	purge_multi_dbh
);

my $dbh;
my $DBS;
my $sub_dbh;

sub __tous_db() {
	return __tous($dbh, 'DB');
}

sub __db_par_name($) {
	my ( $name ) = @_;
	return __par_clef($dbh, 'DB', $name, 'DB_NAME');
}

sub __incr_db_cur($) {
	my ( $dbid ) = @_;
	return __set_tbl_val($dbh, 'DB', 'DB_CUR', dochar($dbid), 'DB_CUR+1');
}

sub __decr_db_cur($) {
	my ( $dbid ) = @_;
	return __set_tbl_val($dbh, 'DB', 'DB_CUR', dochar($dbid), 'DB_CUR-1');
}

sub __new_base($$) {
	my ( $suffix, $dbname ) = @_;
	return __new($dbh, 'BASE',
		'BASE_SUFFIX='.dochar($suffix),
		'BASE_DBNAME='.dochar($dbname)
	);
}

sub __del_base($) {
	my ( $id ) = @_;
	return __del($dbh, 'BASE', $id);
}

sub __base_par_suffix($) {
	my ( $suffix ) = @_;
	return __par_clef($dbh, 'BASE', $suffix, 'BASE_SUFFIX');
}

sub __set_base_version($$) {
	my ( $bid, $version ) = @_;
	return __set_tbl_val($dbh, 'BASE', 'BASE_VERSION', dochar($bid), dochar($version));
}

sub __dispatch($) {
	my ( $suffix ) = @_;
	# Pas besoin de verrouiller pendant le dispatch:
	# - le suffix est deja attribue
	# - l'unicite de l'index sur BASE_SUFFIX garanti que l'ecriture ne peut pas etre en
	#   double dans la table
	my $dbs = [ __tous_db() ];
	if ( ! @$dbs ) {
		die "La table DB de la base principale est vide, je ne peux pas dispatcher un nouveau client. Echec.";
	}
	my $limits = [];
	my $total = 0;
	for my $db ( @$dbs ) {
		$total += $db->{'DB_MAX'};
		$total -= $db->{'DB_CUR'};
		push @$limits, $total;
	}

	my $choice = rand($total);
	while ( $limits->[0] < $choice && $limits->[1] ) {
		shift @$limits;
		shift @$dbs;
	}
	my $dbname = $dbs->[0]->{'DB_NAME'};
	my $bid = __new_base($suffix, $dbname);
	if ( ! $bid ) {
		# Le seul cas d'echec (hors MySql aux fraises) c'est que la ligne a ete
		# ajoutee par un voisin.
		error "Failed to create a line in BASE";
		return __base_par_suffix($suffix);
	}
	__incr_db_cur($dbs->[0]->{'DB_ID'});
	return __base_par_suffix($suffix);
}

sub __switch_db($$) {
	my ( $dbh, $dbname ) = @_;
	$DBS->{'name'} = $dbname;
	create($DBS);
	my $res = sql_transaction($dbh, 'use', 'use '.$dbname);
	if ( !$res->{'ok'} ) {
		error "Failed to switch to database '$dbname' : ".$res->{'err_mess'};
		return undef;
	}
	$dbh->{'dbname'} = $dbname;
}

BEGIN {
	switch_option('dry',0,'all');
	set_quiet('quiet',0);
	set_output('syslog');
}

sub do_connect($) {
	my ( $name ) = @_;
	$dbh = FDN::Sql::DBC::do_connect($name);
	$DBS = read_config("check/$name.cfg","multi");
	read_files($DBS);
	return $dbh;
}

sub reset_multi_dbh {
	return switch_multi_dbh('NONE');
}


sub __switch_multi_dbh($) {
	my ( $suffix ) = @_;
	if ( ! $suffix ) {
		die "Le suffix ne doit jamais etre vide.";
	}
	if ( ! defined $dbh ) {
		die "Il faut se connecter, quand meme, un peu.";
	}
	if ( ! defined $sub_dbh ) {
		$sub_dbh = clone_dbh($dbh);
		$sub_dbh->{'suffix'} //= 'NONE';
		$sub_dbh->{'dbname'} //= 'NONE';
	}
}

sub switch_multi_dbh($) {
	my ( $suffix ) = @_;
	__switch_multi_dbh($suffix);
	if ( $sub_dbh->{'suffix'} eq $suffix ) {
		return $sub_dbh;
	}
	if ( $suffix eq 'NONE' ) {
		$sub_dbh->{'suffix'} = 'NONE';
		return undef;
	}
	if ( $suffix =~ /^[A-Z]/ ) {
		die "Le suffix '$suffix' commence par une lettre capitale, il va se cogner dans le nom de la table. On arrete ici.";
	}
	my $base = __base_par_suffix($suffix);
	if ( ! defined $base ) {
		$base = __dispatch($suffix) or return undef;
	}
	if ( $base->{'BASE_DBNAME'} ne $sub_dbh->{'dbname'} ) {
		__switch_db($sub_dbh, $base->{'BASE_DBNAME'}) or return undef;
	}
	$sub_dbh->{'suffix'} = $suffix;
	if ( $base->{'BASE_VERSION'} ne $DBS->{'signature'} ) {
		# Faut-il verrouiller ?
		# Risque 1: deux updates v1 -> v2 en meme temps
		# - a priori on s'en fiche (certains alter vont echouer, OSEF)
		# Risque 2: risque de flip-flap (web1 vise v2, web2 vise v3), on a v1 en base
		# - les deux veulent monter, en meme temps, et la signature finale (v2 ou v3) ne
		#   sera pas la bonne mais un hybride improbable des deux versions.
		# - c'est un cas qui correspond a un client qui ne se connecte "jamais" et
		#   qui a le bon gout de s'authent pendant que web1 et web2 ne sont pas encore
		#   d'accord sur le numero de version de la base !
		structure($DBS, clone_dbh($sub_dbh)); #FIXME es-ce que ça a réussi ?
		__set_base_version($base->{'BASE_ID'}, $DBS->{'signature'});
	}
	return $sub_dbh;
}

sub purge_multi_dbh($) {
	my ( $suffix ) = @_;
	__switch_multi_dbh($suffix);
	my $base = __base_par_suffix($suffix);
	if ( ! defined $base ) {
		# La base '$suffix' n'existe pas, rien a purger.
		debug "La base '$suffix' n'existe pas.";
		return 1;
	}
	my $db_name = $base->{'BASE_DBNAME'};
	if ( $db_name ne $sub_dbh->{'dbname'} ) {
		__switch_db($sub_dbh, $db_name) or return 0;
	}
	my $db = __db_par_name($db_name);
	if ( ! defined $db ) {
		error "Alors... L'instance '$suffix' en multi-db est rangee dans la base '$db_name' que je ne trouve pas dans la table DB... WTF?";
		return 0;
	}
	$sub_dbh->{'suffix'} = $suffix;
	purge($DBS, clone_dbh($sub_dbh));
	__del_base($base->{'BASE_ID'});


	__decr_db_cur($db->{'DB_ID'});
	return 1;
}


1;
